import 'package:flutter/material.dart';
import 'package:flutter_agnes/src/enums/enums_basic.dart';
import 'package:flutter_agnes/src/widgets/share/none_result.dart';
import 'package:flutter_agnes/src/widgets/share/waiting_result.dart';

typedef Future<T> LoadFuture<T>(int page);
typedef List<dynamic> PageItems<T>(T pageData);
typedef Widget ItemsBuilder(dynamic itemData, int index);
typedef Widget LoadingWidgetBuilder();
typedef int ItemCount<T>(T pageData);
typedef int TotalPages<T>(T pageData);

class ListPaginatedWidget<T> extends StatefulWidget {
  final LoadFuture<T> loadFuture;
  final PageItems<T> pageItems;
  final ItemsBuilder itemsBuilder;
  final LoadingWidgetBuilder loadingWidgetBuilder;
  final ItemCount<T> itemCount;
  final TotalPages<T> totalPages;
  final SliverGridDelegate gridDelegate;
  final ValueChanged<int> onPageChanged;
  final bool pageSnapping;
  final PageController pageController;
  final ListType listType;
  final NoneResult noneResult;

  const ListPaginatedWidget.builder({
    Key key,
    @required this.loadFuture,
    @required this.pageItems,
    @required this.itemsBuilder,
    @required this.itemCount,
    @required this.totalPages,
    this.noneResult,
    this.loadingWidgetBuilder,
    this.listType = ListType.LIST_VIEW,
  })  : this.onPageChanged = null,
        this.pageSnapping = true,
        this.pageController = null,
        this.gridDelegate = null,
        super(key: key);

  @override
  State<StatefulWidget> createState() {
    return ListPaginatedWidgetState<T>(this.loadFuture, this.pageItems, this.itemsBuilder, this.loadingWidgetBuilder, this.itemCount,
        this.gridDelegate, this.onPageChanged, this.pageSnapping, this.pageController, this.listType, this.totalPages);
  }
}

class ListPaginatedWidgetState<T> extends State<ListPaginatedWidget> {
  LoadFuture<T> _loadFuture;
  PageItems<T> _pageItems;
  ItemsBuilder _itemsBuilder;
  LoadingWidgetBuilder _loadingWidgetBuilder;
  ItemCount<T> _itemCount;
  TotalPages<T> _totalPages;
  SliverGridDelegate _gridDelegate;
  ValueChanged<int> _onPageChanged;
  bool _pageSnapping;
  PageController _pageController;
  ListStatus _listStatus;
  ListType _listType;
  List _listItems;
  int _currentPage;
  int _nTotalItems;
  int _nTotalPages;
  T _firstPageData;

  ListPaginatedWidgetState(this._loadFuture, this._pageItems, this._itemsBuilder, this._loadingWidgetBuilder, this._itemCount,
      this._gridDelegate, this._onPageChanged, this._pageSnapping, this._pageController, this._listType, this._totalPages);

  @override
  void initState() {
    _initialize();
    _initialFutureCall();
    super.initState();
  }

  void _initialize() {
    _listStatus = ListStatus.LOADING;
    _listItems = List();
    _currentPage = 0;
    _nTotalItems = 0;
    _firstPageData = null;
  }

  @override
  Widget build(BuildContext context) {
    switch (_listStatus) {
      case ListStatus.SUCCESS:
        return _buildPaginatedWidget();

      case ListStatus.LOADING:
        if (_loadingWidgetBuilder != null)
          return _loadingWidgetBuilder();
        else
          return WaitingResult(label: 'Waiting...', color: Theme.of(context).primaryColor, colorCircular: Theme.of(context).primaryColor);
        break;

      case ListStatus.ERROR:
        if (widget.noneResult != null)
          return widget.noneResult;
        else
          return NoneResult(
              label: 'No results found!', background: Theme.of(context).primaryColor, color: Theme.of(context).primaryColorLight);
        break;

      case ListStatus.EMPTY:
      default:
        if (widget.noneResult != null)
          return widget.noneResult;
        else
          return NoneResult(
              label: 'No results found!', background: Theme.of(context).primaryColor, color: Theme.of(context).primaryColorLight);
        break;
    }
  }

  Widget _buildPaginatedWidget() {
    switch (_listType) {
      case ListType.SLIVERS_LIST:
        return SliverList(
            delegate: SliverChildBuilderDelegate(
          _itemBuilder,
          childCount: _getItemCount(),
        ));

      case ListType.LIST_VIEW:
        return ListView.builder(
          physics: BouncingScrollPhysics(),
          shrinkWrap: true,
          itemCount: _getItemCount(),
          itemBuilder: _itemBuilder,
        );

      default:
        return PageView.builder(
          physics: BouncingScrollPhysics(),
          itemCount: _getItemCount(),
          itemBuilder: _itemBuilder,
          controller: _pageController,
          onPageChanged: _onPageChanged,
          pageSnapping: _pageSnapping,
        );
    }
  }

  Widget _itemBuilder(BuildContext context, int index) {
    if (index < _listItems.length) {
      return _itemsBuilder(_listItems[index], index);
    } else {
      return FutureBuilder(
        future: _loadFuture(_currentPage + 1),
        builder: (BuildContext context, AsyncSnapshot/*<T>*/ snapshot) {
          switch (snapshot.connectionState) {
            case ConnectionState.none:
            case ConnectionState.waiting:
            case ConnectionState.active:
              return _loadingWidgetBuilder();
            case ConnectionState.done:
              _listItems.addAll(_pageItems(snapshot.data));
              if (_currentPage < _nTotalPages) _currentPage++;

              Future.microtask(() {
                setState(() {});
              });
          }
          return _loadingWidgetBuilder();
        },
      );
    }
  }

  int _getItemCount() {
    print('Pagina atual: $_currentPage Total de paginas: $_nTotalPages');
    if (_currentPage < _nTotalPages) {
      return _listItems.length + 1;
    }
    return _listItems.length;
  }

  void changeState({
    LoadFuture<T> pageLoadFuture,
    PageItems<T> pageItemsGetter,
    ItemsBuilder itemsBuilder,
    LoadingWidgetBuilder loadingWidgetBuilder,
    ItemCount<T> totalItemsGetter,
    Key scrollViewKey,
    ScrollPhysics scrollPhysics,
    Axis scrollDirection,
    bool reverse,
    EdgeInsets padding,
    bool shrinkWrap,
    ScrollController scrollController,
    bool primary,
    int semanticChildCount,
    double cacheExtent,
    bool addAutomaticKeepAlives,
    bool addRepaintBoundaries,
    bool addSemanticIndexes,
    double itemExtent,
    SliverGridDelegate gridDelegate,
    ValueChanged<int> onPageChanged,
    bool pageSnapping,
    PageController pageController,
    ListType listType,
    bool resetState = false,
  }) {
    _loadFuture = pageLoadFuture ?? _loadFuture;
    _pageItems = pageItemsGetter ?? _pageItems;
    _itemsBuilder = itemsBuilder ?? _itemsBuilder;
    _loadingWidgetBuilder = loadingWidgetBuilder ?? _loadingWidgetBuilder;
    _itemCount = totalItemsGetter ?? _itemCount;
    _gridDelegate = gridDelegate ?? _gridDelegate;
    _onPageChanged = onPageChanged ?? _onPageChanged;
    _pageSnapping = pageSnapping ?? _pageSnapping;
    _pageController = pageController ?? _pageController;
    _listType = listType ?? _listType;

    if (resetState) {
      _initialize();
      setState(() {});
      _initialFutureCall();
    }
    setState(() {});
  }

  void _initialFutureCall() {
    Future<T> future = _loadFuture(1);
    future.then((T pageData) {
      _firstPageData = pageData;
      _nTotalItems = _itemCount(pageData);
      _nTotalPages = _totalPages(pageData);

      /*if (_pageErrorChecker(pageData)) {
        setState(() {
          _listStatus = ListStatus.ERROR;
        });
      } else*/
      if (_nTotalItems == 0) {
        setState(() {
          _listStatus = ListStatus.EMPTY;
        });
      } else {
        _currentPage++;
        _listItems.addAll(_pageItems(pageData));
        setState(() {
          _listStatus = ListStatus.SUCCESS;
        });
      }
    });
  }
}
